import { assert, zip } from "@canvas-js/utils";
import { MessageId } from "./MessageId.js";
import { MessageSet } from "./MessageSet.js";
import { getAncestorClocks } from "./utils.js";
export class AncestorIndex {
    db;
    static schema = {
        $ancestors: {
            $primary: "key/clock",
            key: "bytes",
            clock: "integer",
            links: "bytes",
        },
    };
    constructor(db) {
        this.db = db;
    }
    async getLinks(messageId, atOrBefore) {
        const delta = messageId.clock - atOrBefore;
        const index = delta <= 0xffffffff ? 0x1f - Math.clz32(delta) : Math.floor(Math.log2(delta));
        const clock = messageId.clock - (1 << index);
        const record = await this.db.get("$ancestors", [messageId.key, clock]);
        if (record === null) {
            throw new Error(`ancestor links not found for ${messageId} at ${clock}`);
        }
        return MessageSet.decode(record.links);
    }
    async isAncestor(messageId, ancestorId, visited) {
        if (messageId.equals(ancestorId)) {
            return true;
        }
        if (messageId.clock <= ancestorId.clock) {
            return false;
        }
        const links = await this.getLinks(messageId, ancestorId.clock);
        for (const link of links) {
            if (visited.has(link)) {
                continue;
            }
            visited.add(link);
            const result = await this.isAncestor(link, ancestorId, visited);
            if (result) {
                return true;
            }
        }
        return false;
    }
    async indexAncestors(messageId, parentIds, effects) {
        messageId = typeof messageId === "string" ? MessageId.encode(messageId) : messageId;
        const ancestorClocks = Array.from(getAncestorClocks(messageId.clock));
        const ancestorLinks = new Array(ancestorClocks.length);
        for (const [i, ancestorClock] of ancestorClocks.entries()) {
            if (i === 0) {
                ancestorLinks[i] = parentIds;
            }
            else {
                const links = new MessageSet();
                for (const child of ancestorLinks[i - 1]) {
                    if (child.clock <= ancestorClock) {
                        links.add(child);
                    }
                    else {
                        assert(child.clock <= ancestorClocks[i - 1], "expected childClock <= ancestorClocks[i - 1]");
                        await this.getAncestors(child, ancestorClock, links);
                    }
                }
                ancestorLinks[i] = links;
            }
        }
        for (const [clock, links] of zip(ancestorClocks, ancestorLinks)) {
            const value = { key: messageId.key, clock, links: links.encode() };
            effects.push({ model: "$ancestors", operation: "set", value });
        }
    }
    async getAncestors(messageId, atOrBefore, results = new MessageSet(), visited = new MessageSet()) {
        assert(atOrBefore > 0, "expected atOrBefore > 0");
        assert(atOrBefore < messageId.clock, "expected atOrBefore < clock");
        const links = await this.getLinks(messageId, atOrBefore);
        for (const ancestor of links) {
            if (ancestor.clock <= atOrBefore) {
                results.add(ancestor);
            }
            else if (visited.has(ancestor)) {
                break;
            }
            else {
                visited.add(ancestor);
                await this.getAncestors(ancestor, atOrBefore, results, visited);
            }
        }
        return results;
    }
}
