// IDs are made by concatenating a **reverse** unsigned varint clock with the hash and then
// truncating to 20 bytes to be base32-friendly, e.g "05vj050kb09l7okead3vvi6so7c7tunn"
import { sha256 } from "@noble/hashes/sha256";
import { base32hex } from "multiformats/bases/base32";
import { encodeClock, decodeClock } from "./clock.js";
export const KEY_LENGTH = 20;
export const ID_LENGTH = 32;
export const MIN_MESSAGE_ID = "00000000000000000000000000000000";
export const MAX_MESSAGE_ID = "vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv";
export const messageIdPattern = /^[0123456789abcdefghijklmnopqrstuv]{32}$/;
export function getKey(clock, value) {
    const hash = sha256(value);
    const key = new Uint8Array(KEY_LENGTH);
    const encodingLength = encodeClock(key, clock);
    key.set(hash.subarray(0, KEY_LENGTH - encodingLength), encodingLength);
    return key;
}
export const encodeId = (id) => base32hex.baseDecode(id);
export const decodeId = (key) => base32hex.baseEncode(key);
export class MessageId {
    id;
    key;
    clock;
    static create(clock, value) {
        const hash = sha256(value);
        const key = new Uint8Array(KEY_LENGTH);
        const encodingLength = encodeClock(key, clock);
        key.set(hash.subarray(0, KEY_LENGTH - encodingLength), encodingLength);
        const id = decodeId(key);
        return new MessageId(id, key, clock);
    }
    static decode(key) {
        const id = decodeId(key);
        const [clock] = decodeClock(key);
        return new MessageId(id, key, clock);
    }
    static encode(id) {
        const key = encodeId(id);
        const [clock] = decodeClock(key);
        return new MessageId(id, key, clock);
    }
    constructor(id, key, clock) {
        this.id = id;
        this.key = key;
        this.clock = clock;
    }
    equals(other) {
        return this.id === other.id;
    }
    toString() {
        return this.id;
    }
}
