import { logger } from "@libp2p/logger";
import { pushable } from "it-pushable";
import * as lp from "it-length-prefixed";
import { pipe } from "it-pipe";
import { assert } from "@canvas-js/utils";
import * as Sync from "@canvas-js/gossiplog/protocols/sync";
import { AbortError } from "@canvas-js/gossiplog/errors";
import { encodeKey, decodeNode } from "./utils.js";
export async function* decodeResponses(source) {
    for await (const msg of source) {
        const res = Sync.Response.decode(msg.subarray());
        yield res;
    }
}
export async function* encodeRequests(source) {
    for await (const req of source) {
        yield Sync.Request.encode(req);
    }
}
export class Client {
    stream;
    responses;
    requests;
    log;
    constructor(stream) {
        this.stream = stream;
        this.log = logger(`canvas:sync:client:[${stream.id}]`);
        this.responses = pipe(stream, lp.decode, decodeResponses);
        this.requests = pushable({ objectMode: true });
        pipe(this.requests, encodeRequests, lp.encode, stream).catch((err) => {
            this.log.error("error piping requests to stream: %O", err);
            if (err instanceof Error) {
                stream.abort(err);
            }
            else {
                stream.abort(new Error("internal error"));
            }
        });
    }
    end() {
        this.log("closing");
        this.requests.end();
    }
    async getRoot() {
        const { getRoot } = await this.get({ getRoot: {} });
        assert(getRoot, "invalid RPC response type");
        assert(getRoot.root !== undefined, "missing `root` in getRoot RPC response");
        return decodeNode(getRoot.root);
    }
    async getNode(level, key) {
        const { getNode } = await this.get({ getNode: { level, key: encodeKey(key) } });
        assert(getNode, "invalid RPC response type");
        if (getNode.node) {
            return decodeNode(getNode.node);
        }
        else {
            return null;
        }
    }
    async getChildren(level, key) {
        const { getChildren } = await this.get({ getChildren: { level, key: encodeKey(key) } });
        assert(getChildren, "invalid RPC response type");
        return getChildren.children.map(decodeNode);
    }
    async getValues(keys) {
        const { getValues } = await this.get({ getValues: { keys } });
        assert(getValues, "invalid RPC response type");
        return getValues.values;
    }
    async get(req) {
        this.log.trace("req: %O", req);
        this.requests.push(req);
        const { done, value: res } = await this.responses.next();
        if (done) {
            this.log.error("stream %s ended prematurely: %O", this.stream.id, res);
            throw new Error("stream ended prematurely");
        }
        this.log.trace("res: %O", res);
        if (res.abort !== undefined) {
            throw new AbortError();
        }
        return res;
    }
}
